import warnings
from typing import Callable
from typing import Sequence


def safely(
        message: str = f"There was a problem a function's application."
    ):
    """
    A decorator for more graceful failing. 
    If the decorated function raises an exception, 
    it will return `None`. 
    

    Args:
        message (str, optional): 
            A warning message in the case of an exception. 
            Defaults to `f"There was a problem a function's application."`.
    """
    def decorator(func:Callable):
        def safe_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                warnings.warn(message)
                return None
        return safe_func
    return decorator

def filter_nones(
    filterer: Sequence,
    to_filter: list[Sequence]
)->list[Sequence]:
    """
    Filter lists based on the presence of None values.

    #### Usage: 
    ```{python}
    # on a single list
    from fasttrackpy.utils.safely import filter_nones
    a = [1, 2, None, 6]

    # value unpacking
    a, = filter_nones(a, [a])
    print(a)
    ```

    ```{python}
    from fasttrackpy.utils.safely import filter_nones
    a = [1, 2, None, 6]
    b = ["a", "b", "c", "d"]

    a,b = filter_nones(a, [a,b])
    print(a)
    print(b)
    ```

    Args:
        filterer (Sequence):
            The filterer list that may contain `None` values
        to_filter (list[Sequence]):
            A list of lists to filter.

    Returns:
        list[Sequence]: _description_
    """
    nones = [i for i,x in enumerate(filterer) if x is None]
    filtered = [
        [x for i,x in enumerate(one_list) if not i in nones]
        for one_list in to_filter
    ]
    return filtered




import numpy as np
from typing import Union
from collections.abc import Callable
class Loss:
    """_summary_

    Args:
        method (Union[str, Callable], optional): _description_. Defaults to "lmse".
    """

    def __init__(
        self,
        method: Union[str, Callable] = "lmse",
        **kwargs
    ):
        self.method = self._get_fun(method)
        self.method_args = kwargs
    
    def _get_fun(
            self, 
            method: Union[str, Callable]
        ) -> Callable:
        if callable(method):
            return method
        if method == "lmse":
            return lmse
        if method == "mse":
            return mse
        
    def calculate_loss(
        self,
        formants: np.ndarray, 
        smoothed: np.ndarray
    ):
        return self.method(formants, smoothed, **self.method_args)
    
def lmse(
        formants: np.ndarray, 
        smoothed: np.ndarray, 
        axis: int = 1
    ) -> np.ndarray:
    """_summary_

    Args:
        formants (np.ndarray): _description_
        smoothed (np.ndarray): _description_
        axis (int, optional): _description_. Defaults to 1.

    Returns:
        np.ndarray: _description_
    """
    sqe = np.power(np.log(formants) - np.log(smoothed), 2)
    mse = np.nanmean(sqe, axis = axis)
    return mse

def mse(
        formants: np.ndarray, 
        smoothed: np.ndarray, 
        axis: int = 1
    ) -> np.ndarray:
    """_summary_

    Args:
        formants (np.ndarray): _description_
        smoothed (np.ndarray): _description_
        axis (int, optional): _description_. Defaults to 1.

    Returns:
        np.ndarray: _description_
    """
    sqe = np.power(formants - smoothed, 2)
    mse = np.nanmean(sqe, axis = axis)
    return mse


import numpy as np
from typing import Union
from collections.abc import Callable

class Agg:
    """_summary_

    Args:
        method (Union[str, Callable], optional): _description_. Defaults to "agg_sum".
    """

    def __init__(
        self, 
        method: Union[str, Callable]= "agg_sum",
        **kwargs
    ):
        self.method = self._get_method(method)
        self.method_args = kwargs
    
    def _get_method(
        self, 
        method:Union[str, Callable]
    ):
        if callable(method):
            return method
        if method == "agg_sum":
            return agg_sum
    
    def aggregate(
        self,
        error
    ):
        return self.method(error, **self.method_args)

def agg_sum(error, axis = 0):
    """
    Sum the error
    """

    agg_error = np.sum(error, axis = axis)
    return agg_error


import numpy as np
import scipy.fft
from typing import Union
from collections.abc import Callable


class Smoothed:
    """Smooothed formant tracks
    Args:
        smoothed (np.ndarray): a (formants, time) shaped numpy array of 
            smoothed formant values
        params (np.ndarray, optional): Parameters (if any) of the smoother. 
            Defaults to None.
    """

    def __init__(
        self,
        smoothed: np.ndarray,
        params: np.ndarray = None
    ):
        self.smoothed = smoothed
        self.params = params

class Smoother:
    """A smoother function factory

    Args:
        method (Union[str, Callable], optional): The smoothing method to use.
            Defaults to "dct_smooth".
            Can be a custom smoother such that it takes a 1D array as input
            and returns a `Smoothed` class.
        kwargs : Any additional arguments or parameters for the `method`.
    """
    def __init__(
        self,
        method: Union[str, Callable] = "dct_smooth_regression",
        **kwargs
    ):
        self.smooth_fun = self._get_fun(method)
        self.method_args = kwargs
    
    def _get_fun(self, method):
        if callable(method):
            return method
        if method == "dct_smooth":
            return dct_smooth
        if method == "dct_smooth_regression":
            return dct_smooth_regression
        
    def smooth(
            self, 
            x: np.array
        ) -> Smoothed:
        """Apply the smoother function to the data

        Args:
            x (np.array): a 1D numpy array

        Returns:
            (Smoothed): A `Smoothed` object
        """
        return self.smooth_fun(x, **self.method_args)

def dct_smooth(
        x:np.array, 
        order:int = 5
    ) -> Smoothed:
    """A DCT Smoother

    Args:
        x (np.array): A 1D array of values to smooth.
        order (int, optional): DCT Order. Defaults to 5.

    Returns:
        (Smoothed): See `Smoothed`
    """
    coefs = scipy.fft.dct(x)
    coef_subset = coefs[0:order]
    smooth = scipy.fft.idct(coef_subset, n = x.shape[0])
    return Smoothed(
        smoothed=smooth, 
        params=coef_subset
    )


def dct_smooth_regression(
        x:np.array, 
        order:int = 5
    ) -> Smoothed:
    """A DCT Smoother using regression

    Args:
        x (np.array): A 1D array to smooth
        order (int, optional): Order of the DCT smoother. 
             Defaults to 5.

    Returns:
        (Smoothed): See `smoothed`
    """

    y = np.array (x)
    N = x.size
    predictors = (scipy.fft.dct(
        np.eye(N), 
        orthogonalize=True,
        norm = "backward"
    )[:,0:order]).T

    nan_entries = np.isnan(y)
    
    y_to_fit = y[~nan_entries]
    predictors_to_use = predictors[:,~nan_entries].T

    try:
        coefs = np.dot(
            (np.linalg.inv(
                np.dot(predictors_to_use.T,
                    predictors_to_use)
                )
            ), 
            np.dot(
                predictors_to_use.T,
                y_to_fit)
            )
        smooth = np.dot(predictors.T, coefs)
    except:
        smooth = np.full(y.shape, np.nan)
        coefs = np.full((order,), np.nan)
    return Smoothed(
        smoothed=smooth,
        params = coefs
    )


import parselmouth as pm
from aligned_textgrid import AlignedTextGrid, Word, Phone, SequenceInterval, SequenceTier
import aligned_textgrid
from fasttrackpy import CandidateTracks, Smoother, Loss, Agg
from fasttrackpy.patterns.just_audio import create_audio_checker
from fasttrackpy.patterns.audio_textgrid import get_interval_classes
from fasttrackpy.utils.safely import safely, filter_nones
import re
from collections import namedtuple
from pathlib import Path
from tqdm import tqdm
from functools import reduce
from operator import add
from joblib import Parallel, cpu_count, delayed
import warnings
import os
import sys


try:
    import magic
    no_magic = False
except:
    warnings.warn("libmagic not found. "\
                "Some audio file types won't be discovered by fasttrack. "\
                "(mp3, ogg, ...)")
    import sndhdr
    from sndhdr import SndHeaders
    no_magic = True

is_audio = create_audio_checker(no_magic=no_magic)

CorpusPair = namedtuple("CorpusPair", field_names=["wav", "tg"])

def get_audio_files(
        corpus_path: str|Path
        )->list[Path]:
    
    if isinstance(corpus_path, str):
        corpus_path = Path(corpus_path)

    all_files = corpus_path.glob("*")
    all_audio = [x for x in all_files if is_audio(x)]
    return all_audio

def get_corpus(
        audio_files:list[Path]
        ) -> list[tuple[Path, Path]]:
    wav_tg = [
        CorpusPair(wav, wav.with_suffix(".TextGrid"))
        for wav in audio_files
        if wav.with_suffix(".TextGrid").exists()
    ]
    return wav_tg

def read_and_associate_tg(
        corpus_pair: CorpusPair,
        entry_classes:list[SequenceInterval] = [Word, Phone]
        ) -> AlignedTextGrid:
    tg = AlignedTextGrid(
        textgrid_path=str(corpus_pair.tg), 
        entry_classes=entry_classes
        )
    setattr(tg, "wav", corpus_pair.wav)
    return tg

def get_target_tiers(
        tg: AlignedTextGrid,
        target_tier: str = "Phone"
    ):

    if all(
        [hasattr(group, target_tier) for group in tg]
    ):
        all_tiers = [getattr(group, target_tier) for group in tg]
        [setattr(x, "wav", x.within.within.wav) for x in all_tiers]
        return all_tiers
    
    tier_names = [tier.name for group in tg for tier in group]
    if target_tier in tier_names:
        all_tiers = [tier for group in tg for tier in group if tier.name == target_tier]
        [setattr(x, "wav", x.within.within.wav) for x in all_tiers]
        return all_tiers
    
    raise Exception(f"Could not {target_tier} target tier in textgrid")

def get_target_intervals(
        target_tiers: list[SequenceTier],
        target_labels:str = "[AEIOU]",
        min_duration = 0.05
        )->list[SequenceInterval]:
    intervals = [
        interval 
        for tier in target_tiers 
        for interval in tier
        if re.match(target_labels, interval.label) and
        (interval.end - interval.start) > min_duration
    ]
    [setattr(interval, "wav", interval.intier.wav) for interval in intervals]

    return intervals

@safely(message = "There was a problem extracting some sounds.")
def get_sound_parts(
        intervals: list[SequenceInterval],
        window_length: float
):
    sound = pm.Sound(str(intervals[0].wav))
    sound_parts = [
        sound.extract_part(from_time = interval.start-(window_length/2), 
        to_time = interval.end+(window_length/2))
        for interval in intervals
    ]

    return sound_parts

@delayed
@safely(message = "There was a problem getting some candidate tracks.")
def get_candidates_delayed(args_dict):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        candidates =  CandidateTracks(**args_dict)
    if candidates.winner.formants.shape[1] == 1:
        warnings.warn("formant tracking error")
    return candidates

@safely(message = "There was a problem getting some candidate tracks.")
def get_candidates(args_dict):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        candidates =  CandidateTracks(**args_dict)
    if candidates.winner.formants.shape[1] == 1:
        warnings.warn("formant tracking error")
    return candidates

def run_candidates(arg_list, parallel:bool):
    if parallel:
        n_jobs = cpu_count()
        all_candidates = Parallel(n_jobs=n_jobs)(
            get_candidates_delayed(args_dict=arg) for arg in tqdm(arg_list)
            )
        return all_candidates
    
    all_candidates = [get_candidates(args_dict=arg) for arg in tqdm(arg_list)]
    return all_candidates

def process_corpus(
        corpus_path: str|Path,
        entry_classes: list = ["Word", "Phone"],
        target_tier: str = "Phone",
        target_labels: str = "[AEIOU]",
        min_duration: float = 0.05,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        smoother: Smoother = Smoother(),
        loss_fun: Loss = Loss(),
        agg_fun: Agg = Agg()
)->list[CandidateTracks]:
    """Given a directory to a corpus of audio/textgrid pairs, return candidates for all vowels.

    Args:
        corpus_path (str | Path): A path to the corpus
        entry_classes (list, optional): Entry classes for the textgrid tiers. 
            Defaults to ["Word", "Phone"].
        target_tier (str, optional): The tier to target. 
            Defaults to "Phone".
        target_labels (str, optional): A regex that will match intervals to target. 
            Defaults to "[AEIOU]".
        min_duration (float, optional): Minimum vowel duration to mention.
            Defaults to 0.05.
        min_max_formant (float, optional): The lowest max-formant value to try. 
            Defaults to 4000.
        max_max_formant (float, optional): The highest max formant to try. 
            Defaults to 7000.
        nstep (int, optional): The number of steps from the min to the max max formant. 
            Defaults to 20.
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis. 
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window. 
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold. 
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use. 
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use. 
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use. 
            Defaults to Agg().

    Returns:
        (list[CandidateTracks]): A list of candidate tracks.
    """
    all_audio = get_audio_files(corpus_path=corpus_path)
    corpus = get_corpus(all_audio)
    entry_classes = get_interval_classes(entry_classes)
    all_tg = [read_and_associate_tg(pair, entry_classes=entry_classes)
              for pair in corpus]
    all_tiers = [get_target_tiers(tg, target_tier=target_tier)
                 for tg in all_tg]
    all_intervals = [
        get_target_intervals(tiers, 
                             target_labels=target_labels, 
                             min_duration=min_duration
                             )
            for tiers in all_tiers
        ]
    all_candidates = []
    for intervals in all_intervals:
        sound_parts = get_sound_parts(intervals, window_length)
        sound_parts, intervals = filter_nones(sound_parts, [sound_parts, intervals])

        arg_list = [
            {
                "samples": x.values,
                "sampling_frequency": x.sampling_frequency,
                "xmin": x.xmin,
                #"interval": interval,
                "min_max_formant": min_max_formant,
                "max_max_formant": max_max_formant,
                "nstep": nstep,
                "n_formants": n_formants,
                "window_length": window_length,
                "time_step" : time_step,
                "pre_emphasis_from": pre_emphasis_from,
                "smoother": smoother,
                "loss_fun":loss_fun,
                "agg_fun": agg_fun
            } for x, interval in zip(sound_parts, intervals)
        ]

        windows_3_12 = os.name != "posix" and \
                sys.version_info.major == 3 and \
                sys.version_info.minor == 12

        candidate_list = run_candidates(
            arg_list, not windows_3_12
        )

        candidate_list, intervals = filter_nones(candidate_list, [candidate_list, intervals])
        for cand, interval in zip(candidate_list, intervals):
            cand.interval = interval
            cand.file_name = Path(str(interval.wav)).stem
        all_candidates += candidate_list
    
    return all_candidates


import parselmouth as pm
import numpy as np
from fasttrackpy.processors.smoothers import Smoother
from fasttrackpy.processors.losses import Loss
from fasttrackpy.processors.aggs import Agg
from fasttrackpy.processors.outputs import formant_to_dataframe,\
                                           param_to_dataframe,\
                                           log_param_to_dataframe,\
                                           get_big_df,\
                                           spectrogram,\
                                           candidate_spectrograms
import matplotlib.pyplot as mp
from aligned_textgrid import SequenceInterval
from aligned_textgrid.sequences.tiers import TierGroup

import polars as pl

from typing import Union, Literal
import warnings
import logging


def _make_candidate(args_dict):
    track = OneTrack(**args_dict)
    return track

class Track:
    """A generic track class to set up attribute values
    """

    def __init__(
            self,
            sound: pm.Sound = None,
            samples: np.array = None,
            sampling_frequency: float = None,
            xmin: float = 0.0,            
            n_formants: int = 4,
            window_length: float = 0.05,
            time_step: float = 0.002,
            pre_emphasis_from: float = 50,
            smoother: Smoother = Smoother(),
            loss_fun: Loss = Loss(),
            agg_fun: Agg = Agg()
    ):
        #self.sound = sound
        if sound:
            self.samples = sound.values
            self.sampling_frequency = sound.sampling_frequency
            self.xmin = sound.xmin
        else:
            self.samples = samples
            self.sampling_frequency = sampling_frequency
            self.xmin = xmin           
        self.n_formants = n_formants
        self.window_length = window_length
        self.time_step = time_step
        self.pre_emphasis_from = pre_emphasis_from
        self.smoother = smoother
        self.loss_fun = loss_fun
        self.agg_fun = agg_fun
        
    @property
    def sound(self):
        sound_obj = pm.Sound(
            self.samples, 
            sampling_frequency = self.sampling_frequency,
            start_time = self.xmin
        )
        return sound_obj        


class OneTrack(Track):
    """A single formant track.

    You can provide *either*

    - A parselmouth `Sound` object to the `sound` argument

    xor

    - An array of audio samples to the `samples` argument
    - The sampling frequency to the `sampling_frequency` argument
    - Any optional time offset to the `xmin` argument.

    If a `Sound` object is passed to `sound`, any values passed to `samples`,
    `sampling_frequency` and `xmin` are ignored.

    Args:
        sound (pm.Sound, optional): A `parselmouth.Sound` object.
        samples (np.ndarray, optional): A numpy array of audio samples.
        sampling_frequency (float, optional): The audio sampling frequency.
        xmin (float, optional): The time offset for the audio. Defaults to 0.0.
        maximum_formant (float): max formant
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis.
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window.
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold.
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use.
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use.
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use.
            Defaults to Agg().

    Attributes:
        maximum_formant (float): The max formant
        time_domain (np.array): The time domain of the formant estimates
        formants (np.ndarray): A (formants, time) array of values. The formants
            as initially estimated by praat-parselmouth
        smoothed_formants (np.ndarray): The smoothed formant values, using
            the method passed to `smoother`.
        parameters (np.ndarray): The smoothing parameters.
        smooth_error (float): The error term between formants and
            smoothed formants.
        file_name (str): The filename of the audio file, if set.
        interval (aligned_textgrid.SequenceInterval): The textgrid interval of the sound, if set.
        id (str): The interval id of the sound, if set.
        group (str): The tier group name of the sound, if set.
    """

    def __init__(
            self,
            maximum_formant: float,
            sound: pm.Sound = None,
            samples: np.array = None,
            sampling_frequency: float = None,
            xmin: float = 0.0,
            n_formants: int = 4,
            window_length: float = 0.025,
            time_step: float = 0.002,
            pre_emphasis_from: float = 50,
            smoother: Smoother = Smoother(),
            loss_fun: Loss = Loss(),
            agg_fun: Agg = Agg()
        ):
        super().__init__(
            sound=sound,
            samples = samples,
            sampling_frequency=sampling_frequency,
            xmin = xmin,
            n_formants=n_formants,
            window_length=window_length,
            time_step=time_step,
            pre_emphasis_from=pre_emphasis_from,
            smoother=smoother,
            loss_fun=loss_fun,
            agg_fun=agg_fun
        )
        self.maximum_formant = maximum_formant

        self.formants, self.time_domain = self._track_formants()
        self.smoothed_list = self._smooth_formants()
        self.smoothed_log_list = self._smooth_log_formants()
        self._file_name = None
        self._id = None
        self._group = None
        self._formant_df = None
        self._param_df = None
        self._log_param_df = None
        self._interval = None

    def __repr__(self):
        return f"A formant track object. {self.formants.shape}"

    def _track_formants(self):
        formant_obj = self.sound.to_formant_burg(
            time_step = self.time_step,
            max_number_of_formants = 5.5,
            maximum_formant = self.maximum_formant,
            window_length = self.window_length,
            pre_emphasis_from = self.pre_emphasis_from
        )

        time_domain = formant_obj.xs()
        tracks = np.array(
            [
                [
                    formant_obj.get_value_at_time(i+1, x)
                    for x in time_domain
                ]
                for i in range(int(np.floor(self.n_formants)))
            ]
        )

        return tracks, time_domain

    def _smooth_formants(self):
        smoothed_list = [
          self.smoother.smooth(x)
            for x in self.formants
        ]
        return smoothed_list
    
    def _smooth_log_formants(self):
        smoothed_list = [
            self.smoother.smooth(x)
            for x in np.log(self.formants)
        ]

        return smoothed_list

    @property
    def smoothed_formants(self):
        return np.array(
            [x.smoothed for x in self.smoothed_list]
        )

    @property
    def parameters(self):
        return np.array(
            [x.params for x in self.smoothed_list]
        )
    
    @property
    def log_parameters(self):
        return np.array([
            x.params for x in self.smoothed_log_list
        ])

    @property
    def smooth_error(self):
        msqe =  self.loss_fun.calculate_loss(
            self.formants[0:self.n_formants],
            self.smoothed_formants[0:self.n_formants]
        )
        error = self.agg_fun.aggregate(msqe)
        return error

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, x):
        self._file_name = x

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, x):
        self._id = x

    @property
    def interval(self):
        return self._interval

    @property
    def group(self):
        return self._group

    @group.setter
    def group(self, groupname):
        self._group = groupname

    @interval.setter
    def interval(self, interval: SequenceInterval):
        self._interval = interval
        self.label = interval.label
        self.id = interval.id
        self.group = self.__get_group(self._interval)

    def __get_group(self, interval):
        if isinstance(interval.within, TierGroup):
            return interval.within.name

        return self.__get_group(interval.within)

    def to_df(
            self,
            output:Literal["formants", "param", "log_param"] = "formants"
            ) -> pl.DataFrame:
        """Output either the formant values or the formant smoothing parameters \
        as a polars dataframe

        Args:
            output (Literal['formants', 'param', 'log_param'], optional): Whether
                to output the formants or the smoothing parameters.
                Defaults to "formants".

        Returns:
            (pl.DataFrame): A `polars.DataFrame`
        """
        if output == "formants"\
              and not isinstance(self._formant_df, pl.DataFrame):
            df =  formant_to_dataframe(self)
            self._formant_df = df
            return df
        if output == "formants":
            return self._formant_df

        if output == "param"\
            and not isinstance(self._param_df, pl.DataFrame):
            df =  param_to_dataframe(self)
            self._param_df = df
            return df
        if output == "param":
            return self._param_df
        
        if output == "log_param"\
            and not isinstance(self._log_param_df, pl.DataFrame):
            df = log_param_to_dataframe(self)
            self._log_param_df = df
            return df
        
        if output == "log_param":
            return self._log_param_df

        raise ValueError("output must be 'formants', 'param', or 'log_param'")

    def spectrogram(self, **kwargs):
        """
        This will plot the spectrogram and formant tracks
        of a single candidate track. If a `file_name` is 
        provided, it will save the plot to disk.

        Args:
            formants (int, optional): 
                The number of formants to plot. 
                Defaults to 3.
            maximum_frequency (float, optional):
                The frequency range the spectrogram and formants
                will be plotted up to. Defaults to 3500.
            tracks (bool, optional): 
                Whether or not to plot the formant tracks. 
                Defaults to True. If False, just the spectogram
                will be plotted.
            dynamic_range (float, optional): 
                A all spectrogram values below the dynamic range. 
                will be plotted as white. Defaults to 60.
            figsize (tuple[float, float], optional):
                Width and height of the figure in inches. 
                Defaults to (8,5).
            color_scale (str, optional): 
                A named matplotlib color scale for the spectrogram.
                Defaults to "Greys". 
                See [here](https://matplotlib.org/stable/users/explain/colors/colormaps.html) 
                for more options.
            file_name (Path | None, optional):
                If provided, how to save the spectrogram. If not provided (None) 
                the plot will show interactively. Defaults to None.
            dpi (float, optional):
                If the plot is being saved, its image resolution in 
                dots per inch. Defaults to 100.
        """
        spectrogram(self, **kwargs)



class CandidateTracks(Track):
    """A class for candidate tracks for a single formant

    You can provide *either*

    - A parselmouth `Sound` object to the `sound` argument

    xor

    - An array of audio samples to the `samples` argument
    - The sampling frequency to the `sampling_frequency` argument
    - Any optional time offset to the `xmin` argument.

    If a `Sound` object is passed to `sound`, any values passed to `samples`,
    `sampling_frequency` and `xmin` are ignored.


    Args:
        sound (pm.Sound, optional): A `parselmouth.Sound` object.
        samples (np.ndarray, optional): A numpy array of audio samples.
        sampling_frequency (float, optional): The audio sampling frequency.
        xmin (float, optional): The time offset for the audio. Defaults to 0.0.
        min_max_formant (float, optional): The lowest max-formant value to try.
            Defaults to 4000.
        max_max_formant (float, optional): The highest max formant to try.
            Defaults to 7000.
        nstep (int, optional): The number of steps from the min to the max max formant.
            Defaults to 20.
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis.
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window.
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold.
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use.
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use.
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use.
            Defaults to Agg().

    Attributes:
        candidates (list[OneTrack,...]): A list of `OneTrack` tracks.
        smooth_errors (np.array): The error terms for each treack in `candidates`
        winner_idx (int): The candidate track with the smallest error term
        winner (OneTrack): The winning `OneTrack` track.
        file_name (str): The filename of the audio file, if set.
        interval (aligned_textgrid.SequenceInterval): The textgrid interval of the sound, if set.
        id (str): The interval id of the sound, if set.
        group (str): The tier group name of the sound, if set.
    """

    def __init__(
        self,
        sound: pm.Sound = None,
        samples: np.array = None,
        sampling_frequency: float = None,
        xmin: float = 0.0,
        min_max_formant: float = 4000,
        max_max_formant: float = 7000,
        nstep = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        smoother: Smoother = Smoother(),
        loss_fun: Loss = Loss(),
        agg_fun: Agg = Agg()
    ):
        super().__init__(
            sound=sound,
            samples = samples,
            sampling_frequency = sampling_frequency,
            xmin = xmin,
            n_formants=n_formants,
            window_length=window_length,
            time_step=time_step,
            pre_emphasis_from=pre_emphasis_from,
            smoother=smoother,
            loss_fun=loss_fun,
            agg_fun=agg_fun
        )

        self.min_max_formant = min_max_formant
        self.max_max_formant = max_max_formant
        self.nstep = nstep
        self.max_formants = np.linspace(
            start = self.min_max_formant,
            stop = self.max_max_formant,
            num = self.nstep
        )
        self._file_name = None
        self._id = None
        self._label = None
        self._group = None
        self._formant_df = None
        self._param_df = None
        self._log_param_df = None
        self._interval = None

        to_process = [
            {
                "samples": self.samples,
                "sampling_frequency": self.sampling_frequency,
                "xmin": self.xmin,
                "maximum_formant": max_formant,
                "n_formants": self.n_formants,
                "window_length": self.window_length,
                "time_step": self.time_step,
                "pre_emphasis_from": self.pre_emphasis_from,
                "smoother": self.smoother,
                "loss_fun": self.loss_fun,
                "agg_fun": self.agg_fun                

            }
            for max_formant in self.max_formants
        ]

        self.candidates = [
            _make_candidate(x) for x in to_process
        ]

        self.smooth_errors = np.array(
            [x.smooth_error for x in self.candidates]
        )

        self.winner_idx = np.argmin(self.smooth_errors)
        self.winner = self.candidates[self.winner_idx]

    @property
    def file_name(self):
        return self._file_name

    @file_name.setter
    def file_name(self, x):
        self._file_name = x
        for c in self.candidates:
            c.file_name = x

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, x):
        self._id = x
        for c in self.candidates:
            c.id = x

    @property
    def group(self):
        return self._group

    @group.setter
    def group(self, name):
        self._group = name

    def __get_group(self, interval):
        if isinstance(interval.within, TierGroup):
            return interval.within.name

        return self.__get_group(interval.within)


    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, interval):
        self._interval = interval
        self.id = interval.id
        self.label = interval.label
        self.group = self.__get_group(interval)
        for c in self.candidates:
            c.interval = interval

    def to_df(
            self,
            which: Literal["winner", "all"] = "winner",
            output: Literal["formants", "param", "log_param"] = "formants"
            ) -> pl.DataFrame:
        """Return a polars dataframe of the candidate tracks

        Args:
            which (Literal['winner', 'all'], optional): Return just the winner
                track data, or all candidates. Defaults to "winner".
            output (Literal['formants', 'param', 'log_param'], optional): Whether
                to output the formants or the smoothing parameters.
                Defaults to "formants".

        Returns:
            (pl.DataFrame): A `polars.DataFrame`
        """
        if which == "winner":
            return self.winner.to_df(output=output)

        if output == "formants"\
            and not isinstance(self._formant_df, pl.DataFrame):
            big_df = get_big_df(self, output=output)
            self._formant_df = big_df
            return big_df

        if output == "formants":
            return self._formant_df

        if output == "param"\
            and not isinstance(self._param_df, pl.DataFrame):
            big_df = get_big_df(self, output=output)
            self._param_df = big_df
            return big_df

        if output == "param":
            return self._param_df
        
        if output == "log_param"\
            and not isinstance(self._log_param_df, pl.DataFrame):
            big_df = get_big_df(self, output=output)
            self._log_param_df = big_df
            return big_df

        if output == "log_param":
            return self._log_param_df        

    def spectrograms(self, **kwargs):
        """ 
        This will plot a grid of the candidate formant
        tracks and their spectrograms. If a `file_name`
        is provided, it will save the plot to disk.

        Args:
            formants (int, optional): 
                The number of formants to plot. 
                Defaults to 3.
            maximum_frequency (float, optional):
                The frequency range the spectrogram and formants
                will be plotted up to. Defaults to 3500.
            dynamic_range (float, optional): 
                A all spectrogram values below the dynamic range. 
                will be plotted as white. Defaults to 60.
            figsize (tuple[float, float], optional):
                Width and height of the figure in inches. 
                Defaults to (8,5).
            file_name (Path | None, optional):
                If provided, how to save the spectrogram. If not provided (None) 
                the plot will show interactively. Defaults to None.
            dpi (float, optional):
                If the plot is being saved, its image resolution in 
                dots per inch. Defaults to 75
        """

        candidate_spectrograms(self, **kwargs)




import parselmouth as pm
from aligned_textgrid import AlignedTextGrid, Word, Phone, SequenceInterval, SequenceTier
import aligned_textgrid
from fasttrackpy import CandidateTracks, Smoother, Loss, Agg
from fasttrackpy.patterns.just_audio import create_audio_checker
from fasttrackpy.utils.safely import safely, filter_nones
import re
import os
import sys

from pathlib import Path
from tqdm import tqdm
from joblib import Parallel, cpu_count, delayed
import warnings

try:
    import magic
    no_magic = False
except:
    warnings.warn("libmagic not found. "\
                "Some audio file types won't be discovered by fasttrack. "\
                "(mp3, ogg, ...)")
    import sndhdr
    from sndhdr import SndHeaders
    no_magic = True

is_audio = create_audio_checker(no_magic=no_magic)

def get_interval_classes(
        textgrid_format: list = ["Word", "Phone"]
        ):
    if all(
        [hasattr(aligned_textgrid, x) for x in textgrid_format]
    ):
        return [getattr(aligned_textgrid, x) for x in textgrid_format]
    
    return [SequenceInterval]

def get_target_tiers(
        tg: AlignedTextGrid,
        target_tier: str = "Phone"
    ):

    if all(
        [hasattr(group, target_tier) for group in tg]
    ):
        return [getattr(group, target_tier) for group in tg]
    
    tier_names = [tier.name for group in tg for tier in group]
    if target_tier in tier_names:
        return [tier for group in tg for tier in group if tier.name == target_tier]
    
    raise Exception(f"Could not {target_tier} target tier in textgrid")

def get_target_intervals(
        target_tiers: list[SequenceTier],
        target_labels:str = "[AEIOU]",
        min_duration = 0.05
        ):
    intervals = [
        interval 
        for tier in target_tiers 
        for interval in tier
        if re.match(target_labels, interval.label) and
        (interval.end - interval.start) > min_duration
    ]

    return intervals

@safely(message="There was a problem getting some candidate tracks.")
def get_candidates(args_dict):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        candidates =  CandidateTracks(**args_dict)
    if candidates.winner.formants.shape[1] == 1:
        warnings.warn("formant tracking error")
    return candidates

@delayed
@safely(message="There was a problem getting some candidate tracks.")
def get_candidates_delayed(args_dict):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        candidates =  CandidateTracks(**args_dict)
    if candidates.winner.formants.shape[1] == 1:
        warnings.warn("formant tracking error")
    return candidates

def run_candidates(arg_list, parallel:bool):
    if parallel:
        n_jobs = cpu_count()
        all_candidates = Parallel(n_jobs=n_jobs)(
            get_candidates_delayed(args_dict=arg) for arg in tqdm(arg_list)
            )
        return all_candidates
    
    all_candidates = [get_candidates(args_dict=arg) for arg in tqdm(arg_list)]
    return all_candidates

def process_audio_textgrid(
        audio_path: str|Path,
        textgrid_path: str|Path,
        entry_classes: list = ["Word", "Phone"],
        target_tier: str = "Phone",
        target_labels: str = "[AEIOU]",
        min_duration: float = 0.05,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        smoother: Smoother = Smoother(),
        loss_fun: Loss = Loss(),
        agg_fun: Agg = Agg()
)->list[CandidateTracks]:
    """Process an audio and TextGrid file together.

    Args:
        audio_path (str | Path): Path to an audio file.
        textgrid_path (str | Path): Path to a TextGrid
        entry_classes (list, optional): Entry classes for the textgrid tiers. 
            Defaults to ["Word", "Phone"].
        target_tier (str, optional): The tier to target. 
            Defaults to "Phone".
        target_labels (str, optional): A regex that will match intervals to target. 
            Defaults to "[AEIOU]".
        min_duration (float, optional): Minimum vowel duration to mention.
            Defaults to 0.05.
        min_max_formant (float, optional): The lowest max-formant value to try. 
            Defaults to 4000.
        max_max_formant (float, optional): The highest max formant to try. 
            Defaults to 7000.
        nstep (int, optional): The number of steps from the min to the max max formant. 
            Defaults to 20.
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis. 
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window. 
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold. 
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use. 
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use. 
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use. 
            Defaults to Agg().

    Returns:
        (list[CandidateTracks]): A list of candidate tracks.
    """
    
    if not is_audio(str(audio_path)):
        raise TypeError(f"The file at {str(audio_path)} is not an audio file")
    
    sound = pm.Sound(str(audio_path))

    entry_classes = get_interval_classes(textgrid_format=entry_classes)
    
    tg = AlignedTextGrid(textgrid_path=textgrid_path, entry_classes=entry_classes)
    target_tiers = get_target_tiers(tg, target_tier=target_tier)
    target_intervals = get_target_intervals(
        target_tiers=target_tiers, 
        target_labels=target_labels,
        min_duration=min_duration
        )
    sound_parts = [
        sound.extract_part(from_time = interval.start-(window_length/2), to_time = interval.end+(window_length/2))
        for interval in target_intervals
    ]

    arg_list = [
        {
            "samples": x.values,
            "sampling_frequency": x.sampling_frequency,
            "xmin": x.xmin,
            #"interval": interval,
            "min_max_formant": min_max_formant,
            "max_max_formant": max_max_formant,
            "nstep": nstep,
            "n_formants": n_formants,
            "window_length": window_length,
            "time_step" : time_step,
            "pre_emphasis_from": pre_emphasis_from,
            "smoother": smoother,
            "loss_fun":loss_fun,
            "agg_fun": agg_fun
        } for x, interval in zip(sound_parts, target_intervals)
    ]
    
    windows_3_12 = os.name != "posix" and \
            sys.version_info.major == 3 and \
            sys.version_info.minor == 12

    candidate_list = run_candidates(
        arg_list, not windows_3_12
    )

    candidate_list, target_intervals = filter_nones(candidate_list, [candidate_list, target_intervals])
    for cand, interval in zip(candidate_list, target_intervals):
        cand.interval = interval
        cand.file_name = Path(str(audio_path)).stem

    return candidate_list





import numpy as np
import pickle
import parselmouth as pm
import polars as pl
from aligned_textgrid import SequenceInterval
from pathlib import Path
import matplotlib.pyplot as mp
import copy
import logging

ptolmap = {"F1" :"#4477AA",
           "F1_s": "#4477AA",
           "F2": "#EE6677",
           "F2_s": "#EE6677",
           "F3": "#228833",
           "F3_s": "#228833",
           "F4": "#CCBB44",
           "F4_s": "#CCBB44"}

def add_metadata(self, out_df):
    if self.file_name:
        out_df = out_df.with_columns(
            file_name = pl.lit(self.file_name)
        )

    if self.id:
        out_df = out_df.with_columns(
            id = pl.lit(self.id)
        )

    if self.group:
        out_df = out_df.with_columns(
            group = pl.lit(self.group)
        )

    if isinstance(self.interval, SequenceInterval) :
        out_df = out_df.with_columns(
            label = pl.lit(self.interval.label)
        )

    return out_df

def formant_to_dataframe(self):
    """Return data as a data frame

    Returns:
        (pl.DataFrame): A data frame
    """
    orig_names = [
        f"F{x}" for x in np.arange(self.n_formants)+1
    ]
    smooth_names = [
        f"F{x}_s" for x in np.arange(self.n_formants)+1
    ]

    orig_df = pl.DataFrame(
        data = self.formants[0:self.n_formants],
        schema=orig_names
    )

    smooth_df = pl.DataFrame(
        data = self.smoothed_formants[0:self.n_formants],
        schema=smooth_names
    )

    out_df = pl.concat([orig_df, smooth_df], how = "horizontal")

    out_df = out_df.with_columns(
        error = pl.lit(self.smooth_error),
        time = pl.lit(self.time_domain),
        max_formant = pl.lit(self.maximum_formant),
        n_formant = pl.lit(self.n_formants),
        smooth_method = pl.lit(self.smoother.smooth_fun.__name__)
    )

    out_df = add_metadata(self, out_df)

    return out_df

def param_to_dataframe(self):
    """Return data as a data frame

    Returns:
        (pl.DataFrame): A data frame
    """

    schema = [
        f"F{x}" for x in
        np.arange(self.parameters.shape[0])+1
    ]
    param_df = pl.DataFrame(
        data = self.parameters,schema=schema
    )

    param_df = param_df.with_columns(
        error = pl.lit(self.smooth_error)
    ).with_row_index(name = "param")

    param_df = add_metadata(self, param_df)

    return param_df

def log_param_to_dataframe(self):
    """Return data as a data frame

    Returns:
        (pl.DataFrame): A data frame
    """

    schema = [
        f"F{x}" for x in
        np.arange(self.log_parameters.shape[0])+1
    ]
    param_df = pl.DataFrame(
        data = self.log_parameters,schema=schema
    )

    param_df = param_df.with_columns(
        error = pl.lit(self.smooth_error)
    ).with_row_index(name = "param")

    param_df = add_metadata(self, param_df)

    return param_df

def get_big_df(self, output):
        all_df = [x.to_df(output = output) for x in self.candidates]
        all_df = [
            x.with_columns(
                candidate = idx+1
            )
            for idx, x in enumerate(all_df)
        ]

        big_df = pl.concat(all_df, how = "diagonal")
        return big_df

def write_data(
        candidates,
        file: Path = None,
        destination: Path = None,
        which: str = "winner",
        output: str = "formants",
        separate: bool = False
):
    if destination and not isinstance(destination, Path):
        destination = Path(destination)

    if file and not isinstance(file, Path):
        file = Path(file)

    if type(candidates) is list:
        df = pl.concat(
            [x.to_df(which = which, output = output) for x in candidates],
            how = "diagonal"
        )
    else:
        df = candidates.to_df(which = which, output = output)

    if file:
        df.write_csv(file = str(file.resolve()))
        return

    if destination and "file_name" in df.columns and not separate:
        file = destination.joinpath(
            df["file_name"][0]
        ).with_suffix(".csv")
        df.write_csv(file = str(file.resolve()))
        return

    if destination and "file_name" in df.columns:
        unique_entries = df \
            .select("file_name", "group")\
            .unique()\
            .with_columns(
                pl.concat_str(
                    [pl.col("file_name"),
                     pl.col("group")],
                     separator="_"
                ).alias("newname")
            ) \
            .rows_by_key("newname", named = True)

        for newname in unique_entries:
            out_df = df \
                .filter(
                    (pl.col("file_name") == unique_entries[newname][0]["file_name"]) &
                    (pl.col("group") == unique_entries[newname][0]["group"])
                )

            out_df.write_csv(
                file = str(destination.joinpath(newname).with_suffix(".csv").resolve())
            )
        return

    if destination:
        file = destination.joinpath("output.csv")
        df.write_csv(file = str(file.resolve()))
        return

    raise ValueError("Either 'file' or 'destination' needs to be set")


def spectrogram(
        self,
        formants:int = 3,
        maximum_frequency:float = 3500,
        tracks:bool = True,
        dynamic_range:float =60,
        figsize: tuple[float, float] = (8,5),
        color_scale: str ="Greys",
        file_name:Path|None = None,
        dpi:float = 100
    ):
    """
    This will plot the spectrogram and formant tracks
    of a single candidate track. If a `file_name` is 
    provided, it will save the plot to disk.

    Args:
        formants (int, optional): 
            The number of formants to plot. 
            Defaults to 3.
        maximum_frequency (float, optional):
            The frequency range the spectrogram and formants
            will be plotted up to. Defaults to 3500.
        tracks (bool, optional): 
            Whether or not to plot the formant tracks. 
            Defaults to True. If False, just the spectogram
            will be plotted.
        dynamic_range (float, optional): 
            A all spectrogram values below the dynamic range. 
            will be plotted as white. Defaults to 60.
        figsize (tuple[float, float], optional):
            Width and height of the figure in inches. 
            Defaults to (8,5).
        color_scale (str, optional): 
            A named matplotlib color scale for the spectrogram.
            Defaults to "Greys". 
            See [here](https://matplotlib.org/stable/users/explain/colors/colormaps.html) 
            for more options.
        file_name (Path | None, optional):
            If provided, how to save the spectrogram. If not provided (None) 
            the plot will show interactively. Defaults to None.
        dpi (float, optional):
            If the plot is being saved, its image resolution in 
            dots per inch. Defaults to 100.
    """

    spctgrm = self.sound.to_spectrogram(
        maximum_frequency=maximum_frequency
    )
    Time, Hz = spctgrm.x_grid(), spctgrm.y_grid()
    db = 10 * np.log10(spctgrm.values)
    min_shown = db.max() - dynamic_range

    mp.figure(figsize=figsize)
    mp.pcolormesh(Time, Hz, db, vmin=min_shown, cmap=color_scale)
    mp.ylim([spctgrm.ymin, spctgrm.ymax])
    mp.xlabel("Time (s)")
    mp.ylabel("Frequency (Hz)")

    formant_cols = [f"F{x+1}" for x in range(formants)]
    smooth_cols = [f"F{x+1}_s" for x in range(formants)]
    data = self.to_df()
    all_cols = [x for x in formant_cols+smooth_cols if x in data.columns]

    data = data\
        .select(["time"]+all_cols)\
        .melt(id_vars = "time")\
        .with_columns(
            pl.col("variable")\
            .replace(ptolmap)\
            .alias("color")
            )

    if tracks:
        mp.scatter(x = "time",
                   y="value",
                   c="color",
                   marker = ".",
                   data=data.filter(
                       ~pl.col("variable").str.contains("_s")
                   ))
        mp.scatter(x = "time",
                   y="value",
                   c="color",
                   marker = "+",
                   data=data.filter(
                       pl.col("variable").str.contains("_s")
                   ))

    if file_name:
        mp.savefig(file_name, dpi=dpi)
        mp.close()


def candidate_spectrograms(
        self,
        formants:int = 3,
        maximum_frequency:float = 3500,
        dynamic_range:float=60,
        figsize:tuple[float, float]=(12,8),
        file_name: Path|None = None,
        dpi:float = 75
    ):
    """ 
    This will plot a grid of the candidate formant
    tracks and their spectrograms. If a `file_name`
    is provided, it will save the plot to disk.

    Args:
        formants (int, optional): 
            The number of formants to plot. 
            Defaults to 3.
        maximum_frequency (float, optional):
            The frequency range the spectrogram and formants
            will be plotted up to. Defaults to 3500.
        dynamic_range (float, optional): 
            A all spectrogram values below the dynamic range. 
            will be plotted as white. Defaults to 60.
        figsize (tuple[float, float], optional):
            Width and height of the figure in inches. 
            Defaults to (8,5).
        file_name (Path | None, optional):
            If provided, how to save the spectrogram. If not provided (None) 
            the plot will show interactively. Defaults to None.
        dpi (float, optional):
            If the plot is being saved, its image resolution in 
            dots per inch. Defaults to 75
    """

    spectrogram = self.sound.to_spectrogram(
        maximum_frequency=maximum_frequency,
        time_step=0.005
        )
    Time = spectrogram.x_grid()
    Hz = spectrogram.y_grid()

    db = 10 * np.log10(spectrogram.values)
    min_shown = db.max() - dynamic_range

    # for plotting layout
    dims = np.array([4, self.nstep//4])
    panel_columns = dims.max()
    panel_rows = dims.min()

    fig = mp.figure(figsize=figsize)
    gs = fig.add_gridspec(panel_rows,panel_columns, hspace=0.18, wspace=0.05)
    axs = gs.subplots(sharex='col', sharey='row')

    formant_cols = [f"F{x+1}" for x in range(formants)]
    smooth_cols = [f"F{x+1}_s" for x in range(formants)]

    for i in range (panel_rows):
        for j in range(panel_columns):
            analysis = i*panel_columns+j

            if analysis == self.winner_idx:
                axs[i, j].pcolormesh(Time, Hz, db, vmin=min_shown, cmap='jet')
            else:
                axs[i, j].pcolormesh(Time, Hz, db, vmin=min_shown, cmap='binary')

            axs[i, j].set_ylim([0, spectrogram.ymax])

            data = self.candidates[analysis].to_df()
            all_cols = [x for x in formant_cols+smooth_cols if x in data.columns]

            data = data\
                .select(["time"]+all_cols)\
                .melt(id_vars = "time")\
                .with_columns(
                    pl.col("variable")\
                    .replace(ptolmap)\
                    .alias("color")
                )

            axs[i,j].scatter(
                x = "time",
                y = "value",
                c = "color",
                data=data.filter(
                       ~pl.col("variable").str.contains("_s")
                ),
                s = 5,
                marker = "."
            )
            axs[i,j].scatter(
                x = "time",
                y = "value",
                c = "color",
                data = data.filter(
                    pl.col("variable").str.contains("_s")
                ),
                s = 5,
                marker = "+"
            )

            axs[i, j].set_title(str(round(self.candidates[analysis].maximum_formant)),y=0.95)

    if file_name:
        mp.savefig(file_name, dpi=dpi, bbox_inches='tight')
        mp.close()

    for ax in fig.get_axes():
        ax.label_outer()


def pickle_candidates(
    candidates,
    file: Path|str
    ):
    """
    This will save a CandidateTracks object to
    disk as a "pickle" file. Due to the way sound 
    objects are handled, only use `unpickle_candidates()`
    to re-load the pickle object.

    Args:
        candidates (CandidateTracks):
            A CandidateTracks object to pickle.
        file (Path | str): 
            The file location to save the pickle file
            to.
    """
    if type(file) is str:
        file = Path(file)

    tmp_candidates = copy.deepcopy (candidates)

    with file.open('wb') as f:
        pickle.dump(tmp_candidates, f)


def unpickle_candidates(
        file: Path|str
        ) :
    """
    This will load a CandidateTracks object 
    that was pickled with `pickle_candidates()`.

    Args:
        file (Path | str):
            The pickled CandidateTracks object
            to unpickle.

    Returns:
        (CandidateTracks):
            A CandidateTracks object.
    """
    if type(file) is str:
        file = Path(file)    

    with file.open('rb') as f:
        candidates = pickle.load(f)

    return candidates



from fasttrackpy.tracks import OneTrack, CandidateTracks
from fasttrackpy.processors.smoothers import Smoother
from fasttrackpy.processors.losses import Loss
from fasttrackpy.processors.aggs import Agg
from fasttrackpy.patterns.just_audio import process_audio_file, process_directory
from fasttrackpy.patterns.audio_textgrid import process_audio_textgrid
from fasttrackpy.patterns.corpus import process_corpus
__all__ = [
    "process_audio_file",
    "process_directory",
    "OneTrack",
    "CandidateTracks",
    "Smoother",
    "Loss",
    "Agg",
    "process_audio_file",
    "process_audio_textgrid",
    "process_corpus"
]


from fasttrackpy.tracks import CandidateTracks
from fasttrackpy.processors.outputs import write_data
from fasttrackpy.processors.aggs import Agg
from fasttrackpy.processors.smoothers import Smoother
from fasttrackpy.processors.losses import Loss
from fasttrackpy.patterns.just_audio import process_audio_file, \
                                            process_directory,\
                                            is_audio
from fasttrackpy.patterns.audio_textgrid import process_audio_textgrid
from fasttrackpy.patterns.corpus import process_corpus
import parselmouth as pm
from pathlib import Path
from typing import Union

import click
import cloup
from cloup import Context, HelpFormatter, HelpTheme, Style
from importlib.resources import files
import yaml

import logging

DEFAULT_CONFIG = str(files("fasttrackpy").joinpath("resources", "config.yml"))

logging.basicConfig(
    filename = "fasttrack.log",
    filemode="a",
    #format="%(message)",
    level=logging.DEBUG
)

def configure(ctx, param, filename):
    with open(filename) as file:
        config = yaml.safe_load(file)
    
    ctx.default_map = config

formatter_settings = HelpFormatter.settings(
    theme=HelpTheme(
        invoked_command=Style(fg='bright_yellow'),
        heading=Style(fg='bright_white', bold=True),
        constraint=Style(fg='magenta'),
        col1=Style(fg='green'),
    )
)

config = cloup.option_group(
    "config file",
    cloup.option(
        "--config",
        type = click.Path(exists=True, dir_okay=False),
        default = DEFAULT_CONFIG,
        callback=configure,
        is_eager=True,
        expose_value=False,
        help="A yaml config file of fasttrack options"
    )
)

audio_processing = cloup.option_group(
    "Audio processing",
    cloup.option(
        "--min-duration",
        type = click.FloatRange(min=0, min_open=True),
        default=0.05,
        help = "Minimum vowel duration"
    ),  
    cloup.option(
        "--min-max-formant", 
        type=click.FloatRange(min=0, min_open=True), 
        default=4000,
        help = "Start of possible max-formant range. Defaults to 4000(Hz)."
    ),
    cloup.option(
        "--max-max-formant", 
        type=click.FloatRange(min=0, min_open=True), 
        default=7000,
        help= "End of possible max-formant range. Defaults to 7000(Hz)."
    ),
    cloup.option(
        "--nstep", 
        type=click.IntRange(min=1), 
        default=20,
        help = "Number of max-formant steps to be evaluated. "\
               "Defaults to 20."
    ),
    cloup.option(
        "--n-formants", 
        type=click.IntRange(min=1), 
        default=4,
        help="Number of formants to track. Defaults to 4."
    ),
    cloup.option(
        "--window-length", 
        type=click.FloatRange(min = 0, min_open=True), 
        default=0.025,
        help = "Formant analysis window length. Defaults to 0.05(s)."
    ),
    cloup.option(
        "--time-step", 
        type=click.FloatRange(min = 0, min_open=True), 
        default=0.002,
        help = "Formant analysis window step size. Defaults to 0.002(s)."
    ),
    cloup.option(
        "--pre-emphasis-from", 
        type=click.FloatRange(min=0), 
        default=50,
        help="Pre-emphasis. Defaults to 50(Hz)."
    )    
)

smoother_options = cloup.option_group(
    "Smoother options",
    cloup.option(
        "--smoother-method", 
        type=click.Choice(["dct_smooth", "dct_smooth_regression"]),
        default="dct_smooth_regression",
        help="Smoother method to use. Defaults to 'dct_smooth_regression' "\
             "(Discrete Cosine Transform)"
    ),
    cloup.option(
        "--smoother-order", 
        type=click.IntRange(min=1), 
        default=5,
        help = "Order of the smooth. Defaults to 5. (More is wigglier.)"
    ),
    cloup.option(
        "--loss-method", 
        type=click.Choice(["lmse", "mse"]), 
        default="lmse",
        help = "The loss function comparing formants to smoothed tracks. "\
               "Defaults to lmse (log mean squared error)."
    )
)

output_options = cloup.option_group(
    "Output Options",
    cloup.option(
        "--which-output", 
        type=click.Choice(["winner", "all"]), 
        default="winner",
        help = "Whether to save just the winner, or all candidates."\
               " Defaults to 'winner'"
    ),
    cloup.option(
        "--data-output", 
        type=click.Choice(["formants", "param", "log_param"]), 
        default="formants",
        help = "Whether to save the formant data, "\
               "or smoothing parameter data."\
               " Defaults to 'formants'."   
    ),
    cloup.option(
        "--separate-output",
        is_flag=True,
        default=False,
        help = "When processing a corpus, save each file/group to a "\
               "separate file?"
    ),
    help = "Options for what data should be saved."
)

textgrid_processing = cloup.option_group(
    "TextGrid Processing",
    cloup.option(
        "--entry-classes",
        type = click.STRING,
        default="Word|Phone",
        help = "Format of the textgrid tier. Default assumes `Word|Phone` forced"\
               " alignment. If not forced alignment, pass `SequenceInterval`"
    ),
    cloup.option(
        "--target-tier",
        type = click.STRING,
        default = "Phone",
        required=True,
        help = "The tier to target. Pass either the entry class (defaults to `Phone`)"\
              " or the tier name."
    ),
    cloup.option(
        "--target-labels",
        type=click.STRING,
        default = "[AEIOU]",
        help="A regex for the intervals to target (defaults to `[AEIOU]`"
    )
)

output_destinations = cloup.option_group(
    "Output Destinations",
    cloup.option(
        "--output", 
        type=click.Path(),
        help = "Name of an output file",
    ),
    cloup.option(
        "--dest", 
        type=click.Path(),
        help = "Name of an output directory"
    ),
    help = "Output destination options",
    constraint=cloup.constraints.RequireAtLeast(1)    
)

@cloup.group(show_subcommand_aliases=True)
def fasttrack():
    """Run fastttrack"""
    pass

@fasttrack.command(
    aliases = ["audio"],
    formatter_settings=formatter_settings,
    help="run fasttrack"
)
@cloup.option_group(
    "Audio Inputs",
    cloup.option(
        "--file", 
        type=click.Path(exists=True),
        help = "A single input audio file to process."
    ),
    cloup.option(
        "--dir", 
        type=click.Path(exists=True),
        help = "A directory of input audio files to process."
    ),
    help="Audio input file options",
    constraint=cloup.constraints.RequireAtLeast(1)
)
@config
@output_destinations
@output_options
@audio_processing
@smoother_options
def audio(
        file: Union[str, Path] = None,
        dir: Union[str,Path] = None,
        output: Union[str, Path] = None,
        dest: Union[str, Path] = None,
        which_output: str = "winner",
        data_output: str = "formants",
        smoother_method: str = "dct_smooth_regression",
        smoother_order: int = 5,
        loss_method: str = "lmse",
        xmin:float = 0,
        xmax: float = None,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        min_duration = 0.05,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        **kwargs
):
    """Run fasttrack.

    Args:
        file (Union[str, Path], optional): A single input audio file to process. 
            Defaults to None.
        dir (Union[str,Path], optional): A directory of input audio files to process. Defaults to None.
        output (Union[str, Path], optional): Name of an output file. Defaults to None.
        dest (Union[str, Path], optional): "Name of an output directory. Defaults to None.
        which_output (str, optional): Whether to save just the winner, 
            or all candidates. Defaults to 'winner'. Defaults to "winner".
        data_output (str, optional): Whether to save the formant data,
            or smoothing parameter data.
            Defaults to "formants".
        smoother_method (str, optional): Smoother method to use. Defaults to 'dct_smooth_regression'
            (Discrete Cosine Transform)
        smoother_order (int, optional): Order of the smooth. 
            Defaults to 5. (More is wigglier.)
        loss_method (str, optional): The loss function comparing formants to 
            smoothed tracks.
            Defaults to lmse (log mean squared error).
        min_max_formant (float, optional): Start of possible max-formant range. 
            Defaults to 4000(Hz).
        max_max_formant (float, optional): End of possible max-formant range. 
            Defaults to 7000(Hz).
        nstep (int, optional): Number of max-formant steps to be evaluated.
            Defaults to 20.
        n_formants (int, optional): Number of formants to track.
            Defaults to 4.
        window_length (float, optional): Formant analysis window length. 
            Defaults to 0.05(s).
        time_step (float, optional): Formant analysis window step size.
            Defaults to 0.002(s)
        pre_emphasis_from (float, optional): Pre-emphasis. Defaults to 50(Hz)
    """
    smoother_kwargs = {
        "method": smoother_method,
        "order": smoother_order
    }

    loss_kwargs = {
        "method": loss_method
    }

    smoother = Smoother(**smoother_kwargs)
    loss_fun = Loss(**loss_kwargs)
    agg_fun = Agg()

    if file and is_audio(file):

        candidates = process_audio_file(
            path = file,
            xmin = 0,
            xmax = None,
            min_max_formant=min_max_formant,
            max_max_formant=max_max_formant,
            nstep=nstep,
            n_formants=n_formants,
            window_length=window_length,
            time_step=time_step,
            pre_emphasis_from=pre_emphasis_from,
            smoother=smoother,
            loss_fun=loss_fun,
            agg_fun=agg_fun
        )

        write_data(candidates=candidates, 
                   file=output, 
                   destination=dest,
                   which=which_output, 
                   output=data_output
        )
    if dir:
        candidate_list = process_directory(
            path = dir,
            min_max_formant=min_max_formant,
            max_max_formant=max_max_formant,
            nstep=nstep,
            n_formants=n_formants,
            window_length=window_length,
            time_step=time_step,
            pre_emphasis_from=pre_emphasis_from,
            smoother=smoother,
            loss_fun=loss_fun,
            agg_fun=agg_fun
        )

        [write_data(
            x, 
            destination=dest,
            which = which_output,
            output=data_output
            ) for x in candidate_list]

@fasttrack.command(
    aliases = ["audio-tg"],
    formatter_settings=formatter_settings,
    help="run fasttrack with audio + textgrid"
)
@cloup.option_group(
    "Inputs",
    cloup.option(
        "--audio",
        type = click.Path(exists=True),
        required=True,
        help = "audio file path"
    ),
    cloup.option(
        "--textgrid",
        type = click.Path(exists=True),
        required = True,
        help = "textgrid file path"
    )
)
@config
@output_destinations
@output_options
@textgrid_processing
@audio_processing
@smoother_options
def audio_textgrid(
        audio: Union[str, Path] = None,
        textgrid: Union[str,Path] = None,
        entry_classes: str = None,
        target_tier: str = None,
        target_labels: str = None,
        output: Union[str, Path] = None,
        dest: Union[str, Path] = None,
        which_output: str = "winner",
        data_output: str = "formants",
        smoother_method: str = "dct_smooth_regression",
        smoother_order: int = 5,
        loss_method: str = "lmse",
        min_duration: float = 0.05,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        **kwargs
):
    """Run fasttrack.

    Args:
        audio (Union[str, Path], optional): A single input audio file to process. 
            Defaults to None.
        textgrid (Union[str, Path], optional): A single input audio file to process. 
            Defaults to None.            
        output (Union[str, Path], optional): Name of an output file. Defaults to None.
        dest (Union[str, Path], optional): "Name of an output directory. Defaults to None.
        which_output (str, optional): Whether to save just the winner, 
            or all candidates. Defaults to 'winner'. Defaults to "winner".
        data_output (str, optional): Whether to save the formant data,
            or smoothing parameter data.
            Defaults to "formants".
        smoother_method (str, optional): Smoother method to use. Defaults to 'dct_smooth_regression'
            (Discrete Cosine Transform)
        smoother_order (int, optional): Order of the smooth. 
            Defaults to 5. (More is wigglier.)
        loss_method (str, optional): The loss function comparing formants to 
            smoothed tracks.
            Defaults to lmse (log mean squared error).
        min_max_formant (float, optional): Start of possible max-formant range. 
            Defaults to 4000(Hz).
        max_max_formant (float, optional): End of possible max-formant range. 
            Defaults to 7000(Hz).
        nstep (int, optional): Number of max-formant steps to be evaluated.
            Defaults to 20.
        n_formants (int, optional): Number of formants to track.
            Defaults to 4.
        window_length (float, optional): Formant analysis window length. 
            Defaults to 0.05(s).
        time_step (float, optional): Formant analysis window step size.
            Defaults to 0.002(s)
        pre_emphasis_from (float, optional): Pre-emphasis. Defaults to 50(Hz)
    """
    smoother_kwargs = {
        "method": smoother_method,
        "order": smoother_order
    }

    loss_kwargs = {
        "method": loss_method
    }

    smoother = Smoother(**smoother_kwargs)
    loss_fun = Loss(**loss_kwargs)
    agg_fun = Agg()

    entry_classes = entry_classes.split("|")

    all_candidates = process_audio_textgrid(
        audio_path=audio,
        textgrid_path=textgrid,
        entry_classes=entry_classes,
        target_tier=target_tier,
        target_labels=target_labels,
        min_duration=min_duration,
        min_max_formant=min_max_formant,
        max_max_formant=max_max_formant,
        nstep=nstep,
        n_formants=n_formants,
        window_length=window_length,
        time_step=time_step,
        pre_emphasis_from=pre_emphasis_from,
        smoother=smoother,
        loss_fun=loss_fun,
        agg_fun=agg_fun
    )

    write_data(candidates=all_candidates, 
                file=output, 
                destination=dest,
                which=which_output, 
                output=data_output
    )

@fasttrack.command(
    formatter_settings=formatter_settings,
    help="run fasttrack with on a corpus directory"
)
@cloup.option_group(
    "Inputs",
    cloup.option(
        "--corpus",
        type = click.Path(exists=True),
        required=True,
        help = "corpus path"
    )
)
@config
@output_destinations
@output_options
@textgrid_processing
@audio_processing
@smoother_options
def corpus(
        corpus: str|Path = None,
        entry_classes: str = None,
        target_tier: str = None,
        target_labels: str = None,
        output: str|Path = None,
        separate_output: bool = False,
        dest: str|Path = None,
        which_output: str = "winner",
        data_output: str = "formants",
        smoother_method: str = "dct_smooth_regression",
        smoother_order: int = 5,
        loss_method: str = "lmse",
        min_duration: float = 0.05,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50
):
    smoother_kwargs = {
        "method": smoother_method,
        "order": smoother_order
    }

    loss_kwargs = {
        "method": loss_method
    }

    smoother = Smoother(**smoother_kwargs)
    loss_fun = Loss(**loss_kwargs)
    agg_fun = Agg()

    entry_classes = entry_classes.split("|")

    all_candidates = process_corpus(
        corpus_path = corpus,
        entry_classes = entry_classes,
        target_tier = target_tier,
        target_labels = target_labels,
        min_duration = min_duration,
        min_max_formant = min_max_formant,
        max_max_formant = max_max_formant,
        nstep = nstep,
        n_formants = n_formants,
        window_length = window_length,
        time_step = time_step,
        pre_emphasis_from = pre_emphasis_from,
        smoother = smoother,
        loss_fun = loss_fun,
        agg_fun = agg_fun
    )

    write_data(
        all_candidates,
        file=output, 
        destination=dest,
        which=which_output, 
        output=data_output,
        separate=separate_output
    )

if __name__ == "__main__":
    fasttrack()


import warnings
from pathlib import Path
from typing import Union
from collections.abc import Callable
import parselmouth as pm
from fasttrackpy import CandidateTracks,\
                        Smoother,\
                        Loss,\
                        Agg
from fasttrackpy.utils.safely import safely, filter_nones

from tqdm import tqdm
from joblib import Parallel, cpu_count, delayed
import os
import sys
import logging
logging.basicConfig(level=logging.INFO)

try:
    import magic
    no_magic = False
except:
    warnings.warn("libmagic not found. "\
                "Some audio file types won't be discovered by fasttrack. "\
                "(mp3, ogg, ...)")
    import sndhdr
    from sndhdr import SndHeaders
    no_magic = True

def create_audio_checker(no_magic:bool = no_magic) -> Callable:
    """Return an audio checker, dependent on 
       availability of libmagic.

    Args:
        no_magic (bool): is libmagic available

    Returns:
        (Callable): A sound file checker
    """

    def magic_checker(path: str)->bool:
        """Checks whether a file is an audio file using libmagic

        Args:
            path (str): Path to the file in question

        Returns:
            (bool): Whether or not the file is an audio file
        """
        file_mime = magic.from_file(str(path), mime=True)
        return "audio" in file_mime
    
    def sndhdr_checker(path: str)->bool:
        """Checks whether a file is an audio file using `sndhdr`

        Args:
            path (str): Path to the file

        Returns:
            (bool): Whether or not the file is an audio file.
        """
        hdr_info = sndhdr.what(str(path))
        return isinstance(hdr_info, SndHeaders)
    
    if no_magic:
        return sndhdr_checker
    
    return magic_checker

is_audio = create_audio_checker(no_magic=no_magic)

def process_audio_file(
        path: str|Path,
        xmin:float = 0,
        xmax: float = None,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.025,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        smoother: Smoother = Smoother(),
        loss_fun: Loss = Loss(),
        agg_fun: Agg = Agg()
)->CandidateTracks:
    """Given the path to a single audio file, return a candidates track object.

    Args:
        path (str|Path): Path to the audio file
        xmin (float, optional): Start time to process the audio. Defaults to 0.
        xmax (float, optional): End tome for processing audio. If None, defaults to the
            maximum time. Defaults to None.
        min_max_formant (float, optional): The lowest max-formant value to try. 
            Defaults to 4000.
        max_max_formant (float, optional): The highest max formant to try. 
            Defaults to 7000.
        nstep (int, optional): The number of steps from the min to the max max formant. 
            Defaults to 20.
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis. 
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window. 
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold. 
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use. 
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use. 
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use. 
            Defaults to Agg().

    Returns:
        (CandidateTracks): A `CandidateTracks` object to use.
    """
    if not is_audio(str(path)):
        raise TypeError(f"The file at {str(path)} is not an audio file")
    
    sound = pm.Sound(str(path))
    if not xmax:
        xmax = sound.xmax

    sound_to_process = sound.extract_part(from_time = xmin, to_time = xmax)
    candidates = CandidateTracks(
        samples=sound_to_process.values,
        sampling_frequency=sound_to_process.sampling_frequency,
        xmin = sound_to_process.xmin,
        min_max_formant=min_max_formant,
        max_max_formant=max_max_formant,
        nstep=nstep,
        n_formants=n_formants,
        window_length=window_length,
        time_step=time_step,
        pre_emphasis_from=pre_emphasis_from,
        smoother=smoother,
        loss_fun=loss_fun,
        agg_fun=agg_fun
    )
    candidates.file_name = Path(str(path)).name
    return candidates

@delayed
@safely(message = "There was a problem processing an audio file.")
def get_candidates_delayed(args_dict):
    return process_audio_file(**args_dict)

@safely(message = "There was a problem processing an audio file.")
def get_candidates(args_dict):
    return process_audio_file(**args_dict)

def run_candidates(arg_list, parallel:bool):
    if parallel:
        n_jobs = cpu_count()
        all_candidates = Parallel(n_jobs=n_jobs)(
            get_candidates_delayed(args_dict=arg) for arg in tqdm(arg_list)
            )
        return all_candidates
    
    all_candidates = [get_candidates(args_dict=arg) for arg in tqdm(arg_list)]
    return all_candidates

def process_directory(
        path: str|Path,
        min_max_formant:float = 4000,
        max_max_formant:float = 7000,
        nstep:int = 20,
        n_formants: int = 4,
        window_length: float = 0.05,
        time_step: float = 0.002,
        pre_emphasis_from: float = 50,
        smoother: Smoother = Smoother(),
        loss_fun: Loss = Loss(),
        agg_fun: Agg = Agg()
)->list[CandidateTracks]:
    """Given a path to a directoy of audio files, process them all.

    Args:
        path (str|Path): Path to the directory to process.
        min_max_formant (float, optional): The lowest max-formant value to try. 
            Defaults to 4000.
        max_max_formant (float, optional): The highest max formant to try. 
            Defaults to 7000.
        nstep (int, optional): The number of steps from the min to the max max formant. 
            Defaults to 20.
        n_formants (int, optional): The number of formants to track. Defaults to 4.
        window_length (float, optional): Window length of the formant analysis. 
            Defaults to 0.025.
        time_step (float, optional): Time step of the formant analyusis window. 
            Defaults to 0.002.
        pre_emphasis_from (float, optional): Pre-emphasis threshold. 
            Defaults to 50.
        smoother (Smoother, optional): The smoother method to use. 
            Defaults to `Smoother()`.
        loss_fun (Loss, optional): The loss function to use. 
            Defaults to Loss().
        agg_fun (Agg, optional): The loss aggregation function to use. 
            Defaults to Agg().

    Returns:
        (list[CandidateTracks]): A list of `CandidateTracks` objects.
    """
    if not isinstance(path, Path) and isinstance(path, str):
        path = Path(path)

    all_files = list(path.glob("*"))
    all_files = [x for x in all_files if x.is_file()]
    all_audio = [x for x in all_files if is_audio(x)]
    arg_list = [
            {"path": x,
            "min_max_formant": min_max_formant,
            "max_max_formant":max_max_formant,
            "nstep": nstep,
            "n_formants": n_formants,
            "window_length":window_length,
            "time_step":time_step,
            "pre_emphasis_from":pre_emphasis_from,
            "smoother":smoother,
            "loss_fun":loss_fun,
            "agg_fun":agg_fun
            }
            for x in all_audio
    ]

    windows_3_12 = os.name != "posix" and \
            sys.version_info.major == 3 and \
            sys.version_info.minor == 12

    all_candidates = run_candidates(
        arg_list, not windows_3_12
    )

    all_candidates, all_audio = filter_nones(all_candidates, [all_candidates, all_audio])

    for x, path in zip(all_candidates, all_audio):
        x.file_name = Path(str(path)).name

    return all_candidates




# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fasttrackpy',
 'fasttrackpy.patterns',
 'fasttrackpy.processors',
 'fasttrackpy.utils']

package_data = \
{'': ['*'], 'fasttrackpy': ['resources/*']}

install_requires = \
['aligned-textgrid>=0.6.2,<0.7.0',
 'click>=8.1.7,<9.0.0',
 'cloup>=3.0.3,<4.0.0',
 'joblib>=1.3.2,<2.0.0',
 'matplotlib>=3.8.2,<4.0.0',
 'polars>=0.20.18,<0.21.0',
 'praat-parselmouth>=0.4.3,<0.5.0',
 'pytest-cov>=4.1.0,<5.0.0',
 'pytest>=7.4.3,<8.0.0',
 'pyyaml>=6.0.1,<7.0.0',
 'tqdm>=4.66.1,<5.0.0']

extras_require = \
{':python_version >= "3.10" and python_version < "3.13"': ['scipy>=1.11.3,<2.0.0',
                                                           'numpy>=1.26.1,<2.0.0'],
 ':sys_platform != "win32"': ['python-magic>=0.4.27,<0.5.0'],
 ':sys_platform == "win32"': ['python-magic-bin>=0.4.14,<0.5.0']}

entry_points = \
{'console_scripts': ['fasttrack = fasttrackpy.cli:fasttrack']}

setup_kwargs = {
    'name': 'fasttrackpy',
    'version': '0.4.6',
    'description': 'A python implementation of FastTrack',
    'long_description': '# FastTrackPy\n[![PyPI](https://img.shields.io/pypi/v/fasttrackpy)](https://pypi.org/project/fasttrackpy/) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/fasttrackpy) ![PyPI - Python Version](https://img.shields.io/pypi/pyversions/fasttrackpy) [![Python CI](https://github.com/JoFrhwld/fasttrackpy/actions/workflows/test-and-run.yml/badge.svg)](https://github.com/JoFrhwld/fasttrackpy/actions/workflows/test-and-run.yml) [![codecov](https://codecov.io/gh/FastTrackiverse/fasttrackpy/graph/badge.svg?token=GOAWY4B5C8)](https://codecov.io/gh/FastTrackiverse/fasttrackpy) <a href="https://codeclimate.com/github/JoFrhwld/fasttrackpy/maintainability"><img src="https://api.codeclimate.com/v1/badges/6725fded174b21a3c59f/maintainability" /></a> [![DOI](https://zenodo.org/badge/580169086.svg)](https://zenodo.org/badge/latestdoi/580169086)\n\n\nA python implementation of the FastTrack method\n\n## Installation\n\n```bash\npip install fasttrackpy\n```\n\nThis will make the command line executable `fasttrack` available, along with its subcommands:\n\n- `audio`\n- `audio-textgrid`\n- `corpus`\n\n## Getting help\n\nFor any of the fasttrack subcommands, add the `--help` flag to\nprint the help info. You can also visit [the docs](https://fasttrackiverse.github.io/fasttrackpy/usage/getting_started.html).\n\n## Usage\n\nFor a single audio file containing a vowel-like sound:\n\n```bash\nfasttrack audio --file audio.wav \\\n    --output formants.csv\n```\n\nFor a paired audio file and textgrid with intervals defining\ntarget audio to process:\n\n```bash\nfasttrack audio-textgrid --audio audio.wav \\\n    --textgrid audio.TextGrid \\\n    --output formants.csv\n```\n\nFor a corpus directory of paired audio files and textgrid\n\n```bash\nfasttrack corpus --corpus dir/ \\\n    --output formants.csv\n```',
    'author': 'JoFrhwld',
    'author_email': 'JoFrhwld@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://fasttrackiverse.github.io/fasttrackpy/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.13',
}


setup(**setup_kwargs)
